"""
Test module for the calibration module of the duet_tools package.
"""

from __future__ import annotations

import pytest
import numpy as np
from pathlib import Path
import math
import matplotlib.pyplot as plt

from duet_tools.calibration import (
    DuetRun,
    Targets,
    FuelParameter,
    import_duet,
    import_duet_manual,
    assign_targets,
    set_fuel_parameter,
    set_loading,
    set_moisture,
    set_depth,
    calibrate,
    _maxmin_calibration,
)

from duet_tools.utils import read_dat_to_array

TEST_DIR = Path(__file__).parent
TMP_DIR = TEST_DIR / "tmp"
TMP_DIR.mkdir(exist_ok=True)
DATA_DIR = TEST_DIR / "test-data"


class TestDuetRun:
    def test_import_duet_v1(self):
        duet_run = import_duet_manual(
            directory=DATA_DIR / "v1",
            loading_grid_name="surface_rhof.dat",
            moisture_grid_name="surface_moist.dat",
            depth_grid_name="surface_depth.dat",
            nx=976,
            ny=1998,
            nsp=2,
            version="v1",
        )
        # test that data types are correct
        assert isinstance(duet_run, DuetRun)
        assert isinstance(duet_run.loading, np.ndarray)
        assert isinstance(duet_run.moisture, np.ndarray)
        assert isinstance(duet_run.depth, np.ndarray)
        # test array shapes
        assert duet_run.loading.shape == (2, 1998, 976)
        assert duet_run.moisture.shape == (2, 1998, 976)
        assert duet_run.depth.shape == (2, 1998, 976)
        # test that wrong dimensions raises error
        with pytest.raises(ValueError):
            duet_run = import_duet_manual(
                directory=DATA_DIR / "v1",
                loading_grid_name="surface_rhof.dat",
                moisture_grid_name="surface_moist.dat",
                depth_grid_name="surface_depth.dat",
                nx=976,
                ny=1998,
                nsp=3,
                version="v1",
            )
        # test that wrong version number raises error
        with pytest.raises(ValueError):
            duet_run = import_duet_manual(
                directory=DATA_DIR / "v1",
                loading_grid_name="surface_rhof.dat",
                moisture_grid_name="surface_moist.dat",
                depth_grid_name="surface_depth.dat",
                nx=976,
                ny=1998,
                nsp=2,
                version="v3",
            )

    def test_import_duet_v2(self):
        duet_run = import_duet(directory=DATA_DIR / "v2")
        assert isinstance(duet_run, DuetRun)
        # test that data types are correct
        assert isinstance(duet_run.loading, np.ndarray)
        assert isinstance(duet_run.moisture, np.ndarray)
        assert isinstance(duet_run.depth, np.ndarray)
        # test array shapes
        assert duet_run.loading.shape == (3, 295, 333)
        assert duet_run.moisture.shape == (3, 295, 333)
        assert duet_run.depth.shape == (3, 295, 333)

    def test_import_duet_v2_manual(self):
        duet_run = import_duet_manual(
            directory=DATA_DIR / "v2",
            loading_grid_name="surface_rhof_layered.dat",
            moisture_grid_name="surface_moist_layered.dat",
            depth_grid_name="surface_depth_layered.dat",
            nx=333,
            ny=295,
            nsp=9,
            version="v2",
        )
        assert isinstance(duet_run, DuetRun)
        # test that data types are correct
        assert isinstance(duet_run.loading, np.ndarray)
        assert isinstance(duet_run.moisture, np.ndarray)
        assert isinstance(duet_run.depth, np.ndarray)
        # test array shapes
        assert duet_run.loading.shape == (3, 295, 333)
        assert duet_run.moisture.shape == (3, 295, 333)
        assert duet_run.depth.shape == (3, 295, 333)
        # test that wrong dimensions raise error
        with pytest.raises(ValueError):
            duet_run = import_duet_manual(
                directory=DATA_DIR / "v2",
                loading_grid_name="surface_rhof.dat",
                moisture_grid_name="surface_moist.dat",
                depth_grid_name="surface_depth.dat",
                nx=333,
                ny=295,
                nsp=3,
                version="v2",
            )

    def test_to_numpy(self):
        duet_run = import_duet(directory=DATA_DIR / "v2")
        # test each fuel parameter and type
        grass_loading = duet_run.to_numpy("grass", "loading")
        coniferous_loading = duet_run.to_numpy("coniferous", "loading")
        deciduous_loading = duet_run.to_numpy("deciduous", "loading")
        litter_loading = duet_run.to_numpy("litter", "loading")
        grass_moisture = duet_run.to_numpy("grass", "moisture")
        coniferous_moisture = duet_run.to_numpy("coniferous", "moisture")
        deciduous_moisture = duet_run.to_numpy("deciduous", "moisture")
        litter_moisture = duet_run.to_numpy("litter", "moisture")
        grass_depth = duet_run.to_numpy("grass", "depth")
        coniferous_depth = duet_run.to_numpy("coniferous", "depth")
        deciduous_depth = duet_run.to_numpy("deciduous", "depth")
        litter_depth = duet_run.to_numpy("litter", "depth")
        assert np.array_equal(grass_loading, duet_run.loading[0, :, :])
        assert np.array_equal(coniferous_loading, duet_run.loading[1, :, :])
        assert np.array_equal(deciduous_loading, duet_run.loading[2, :, :])
        assert np.array_equal(
            litter_loading, np.sum(duet_run.loading[1:, :, :], axis=0)
        )
        assert np.array_equal(grass_moisture, duet_run.moisture[0, :, :])
        assert np.array_equal(coniferous_moisture, duet_run.moisture[1, :, :])
        assert np.array_equal(deciduous_moisture, duet_run.moisture[2, :, :])
        litter_weights = _maxmin_calibration(duet_run.loading[1:, :, :], max=1.0, min=0)
        litter_weights[litter_weights == 0] = 0.01
        litter_masked = np.ma.masked_array(
            duet_run.moisture[1:, :, :], duet_run.moisture[1:, :, :] == 0
        )
        litter_averaged = np.ma.average(litter_masked, axis=0, weights=litter_weights)
        weighted_average_litter_moisture = np.ma.filled(litter_averaged, 0)
        assert np.array_equal(litter_moisture, weighted_average_litter_moisture)
        assert np.array_equal(grass_depth, duet_run.depth[0, :, :])
        assert np.array_equal(coniferous_depth, duet_run.depth[1, :, :])
        assert np.array_equal(deciduous_depth, duet_run.depth[2, :, :])
        assert np.array_equal(litter_depth, np.sum(duet_run.depth[1:, :, :], axis=0))
        # test integrated and separated
        separated_loading = duet_run.to_numpy("separated", "loading")
        assert np.array_equal(separated_loading, duet_run.loading)
        integrated_loading = duet_run.to_numpy("integrated", "loading")
        integrated_depth = duet_run.to_numpy("integrated", "depth")
        integrated_moisture = duet_run.to_numpy("integrated", "moisture")
        assert np.array_equal(integrated_loading, np.sum(duet_run.loading, axis=0))
        assert np.array_equal(integrated_depth, np.max(duet_run.depth, axis=0))
        weights = _maxmin_calibration(duet_run.loading.copy(), max=1.0, min=0)
        weights[weights == 0] = 0.01
        masked = np.ma.masked_array(duet_run.moisture, duet_run.moisture == 0)
        averaged = np.ma.average(masked, axis=0, weights=weights)
        weighted_average_moisture = np.ma.filled(averaged, 0)
        assert np.array_equal(integrated_moisture, weighted_average_moisture)

    def test_to_quicfire(self):
        duet_run = import_duet(directory=DATA_DIR / "v2")
        duet_run.to_quicfire(TMP_DIR, overwrite=True)
        with pytest.raises(FileExistsError):
            duet_run.to_quicfire(TMP_DIR)
        duet_run.to_quicfire(TMP_DIR, overwrite=True)
        treesrhof = read_dat_to_array(TMP_DIR, "treesrhof.dat", nx=333, ny=295, nz=1)
        treesrhof = treesrhof[0, :, :]
        treesmoist = read_dat_to_array(
            TMP_DIR,
            "treesmoist.dat",
            nx=333,
            ny=295,
            nz=1,
        )
        treesmoist = treesmoist[0, :, :]
        treesfueldepth = read_dat_to_array(
            TMP_DIR,
            "treesfueldepth.dat",
            nx=333,
            ny=295,
            nz=1,
        )
        treesfueldepth = treesfueldepth[0, :, :]
        assert np.array_equal(
            treesrhof, duet_run._integrate_all("loading").astype("float32")
        )
        assert np.array_equal(
            treesfueldepth, duet_run._integrate_all("depth").astype("float32")
        )
        assert np.array_equal(
            treesmoist, duet_run._integrate_all("moisture").astype("float32")
        )

        files = ["treesrhof.dat", "treesmoist.dat", "treesfueldepth.dat"]
        for file in files:
            path = TMP_DIR / file
            path.unlink()
        duet_run.to_quicfire(TMP_DIR, loading=False, moisture=False, depth=False)
        for file in files:
            path = TMP_DIR / file
            assert path.exists() == False


class TestAssignTargets:
    def test_assign_targets(self):
        maxmin_targets = assign_targets(method="maxmin", max=1.0, min=0.2)
        meansd_targets = assign_targets(method="meansd", mean=0.6, sd=0.03)
        constant_target = assign_targets(method="constant", value=1.0)
        assert isinstance(maxmin_targets, Targets)
        assert isinstance(meansd_targets, Targets)
        assert isinstance(constant_target, Targets)
        assert maxmin_targets.targets == [1.0, 0.2]
        assert meansd_targets.targets == [0.6, 0.03]
        assert constant_target.targets == [1.0]
        # test other attributes
        assert maxmin_targets.args == ["max", "min"]
        assert maxmin_targets.method == "maxmin"
        assert maxmin_targets.calibration_function == _maxmin_calibration

    def test_target_validation(self):
        # test wrong method
        with pytest.raises(ValueError):
            assign_targets(method="minmax", min=0.2, max=1.0)
        # test wrong kwargs
        with pytest.raises(ValueError):
            assign_targets(method="maxmin", max=1.0)
        with pytest.raises(ValueError):
            assign_targets(method="constant", max=1.0, min=0.2)
        with pytest.raises(ValueError):
            assign_targets(method="meansd", max=1.0, min=0.02)
        # test incorrect inputs
        with pytest.raises(ValueError):
            assign_targets(method="maxmin", max=0.2, min=1.0)
        with pytest.warns(UserWarning):
            assign_targets(method="meansd", mean=0.03, sd=0.6)


class TestSetFuelParameter:
    def test_set_fuel_parameter(self):
        maxmin_targets = assign_targets(method="maxmin", max=1.0, min=0.2)
        meansd_targets = assign_targets(method="meansd", mean=0.6, sd=0.03)
        constant_target = assign_targets(method="constant", value=1.0)
        # test separated functions
        loading_targets = set_loading(grass=maxmin_targets, litter=meansd_targets)
        moisture_targets = set_moisture(grass=constant_target)
        depth_targets = set_depth(all=maxmin_targets)
        assert isinstance(loading_targets, FuelParameter)
        assert isinstance(moisture_targets, FuelParameter)
        assert isinstance(depth_targets, FuelParameter)
        assert loading_targets.parameter == "loading"
        assert moisture_targets.parameter == "moisture"
        # test generic function
        loading_targets = set_fuel_parameter(
            parameter="loading", grass=maxmin_targets, litter=meansd_targets
        )
        assert isinstance(loading_targets, FuelParameter)
        assert loading_targets.parameter == "loading"
        assert loading_targets.fuel_types == ["grass", "litter"]
        assert loading_targets.targets == [maxmin_targets, meansd_targets]
        # test validation
        with pytest.raises(ValueError):
            loading_targets = set_fuel_parameter(
                parameter="bulk loading", litter=meansd_targets
            )
        with pytest.raises(ValueError):
            loading_targets = set_fuel_parameter(
                parameter="loading", both=meansd_targets
            )
        with pytest.raises(ValueError):
            loading_targets = set_fuel_parameter(
                parameter="loading", litter=meansd_targets, all=constant_target
            )


class TestCalibrate:
    def test_calibrate_maxmin(self):
        # try 1 fueltype and 1 parameter
        duet_run = import_duet(DATA_DIR / "v2")
        grass_loading = assign_targets(method="maxmin", max=1.0, min=0.2)
        loading_targets = set_fuel_parameter(parameter="loading", grass=grass_loading)
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=loading_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.loading, np.ndarray)
        assert np.allclose(calibrated_duet.depth, duet_run.depth)
        assert np.allclose(calibrated_duet.loading, duet_run.loading) == False
        assert np.allclose(calibrated_duet.loading[1, :, :], duet_run.loading[1, :, :])
        assert np.allclose(calibrated_duet.loading[2, :, :], duet_run.loading[2, :, :])
        assert np.max(calibrated_duet.loading[0, :, :]) == 1.0
        assert np.min(calibrated_duet.loading[0, :, :]) == 0.2
        # try loading and depth
        grass_depth = assign_targets(method="maxmin", max=1.0, min=0.2)
        loading_targets = set_fuel_parameter(parameter="loading", grass=grass_loading)
        depth_targets = set_fuel_parameter(parameter="depth", grass=grass_depth)
        calibrated_duet = calibrate(
            duet_run, fuel_parameter_targets=[loading_targets, depth_targets]
        )
        # now do moisture
        grass_moisture = assign_targets(method="maxmin", max=0.5, min=0.05)
        moisture_targets = set_fuel_parameter(
            parameter="moisture", grass=grass_moisture
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=moisture_targets)

        # try with two fueltypes
        litter_loading = assign_targets(method="maxmin", max=0.1, min=0.01)
        loading_targets = set_fuel_parameter(
            parameter="loading", litter=litter_loading, grass=grass_loading
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=loading_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.loading, np.ndarray)
        assert np.allclose(calibrated_duet.depth, duet_run.depth)
        assert np.allclose(calibrated_duet.loading, duet_run.loading) == False
        assert np.max(calibrated_duet.loading[0, :, :]) == 1.0
        assert np.max(calibrated_duet.loading[1:, :, :]) == 0.1

        # try with three fueltypes
        coniferous_loading = assign_targets(method="maxmin", max=0.1, min=0.01)
        deciduous_loading = assign_targets(method="maxmin", max=0.5, min=0.2)
        loading_targets = set_fuel_parameter(
            parameter="loading",
            coniferous=coniferous_loading,
            deciduous=deciduous_loading,
            grass=grass_loading,
        )
        # there is no deciduous litter, so it will throw an error
        with pytest.raises(ValueError):
            calibrated_duet = calibrate(
                duet_run, fuel_parameter_targets=loading_targets
            )
        # now without deciduous litter targets
        loading_targets = set_fuel_parameter(
            parameter="loading",
            coniferous=coniferous_loading,
            grass=grass_loading,
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=loading_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.loading, np.ndarray)
        assert np.allclose(calibrated_duet.depth, duet_run.depth)
        assert np.allclose(calibrated_duet.loading, duet_run.loading) == False
        assert np.max(calibrated_duet.loading[0, :, :]) == 1.0
        assert np.max(calibrated_duet.loading[1, :, :]) == 0.1
        assert np.max(calibrated_duet.loading[2, :, :]) == 0

    def test_calibrate_meansd(self):
        # try 1 fueltype and 1 parameter
        duet_run = import_duet(DATA_DIR / "v2")
        grass_loading = assign_targets(method="meansd", mean=0.6, sd=0.3)
        loading_targets = set_fuel_parameter(parameter="loading", grass=grass_loading)
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=loading_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.loading, np.ndarray)
        assert np.allclose(calibrated_duet.depth, duet_run.depth)
        assert np.allclose(calibrated_duet.loading, duet_run.loading) == False
        assert np.allclose(
            calibrated_duet.loading[1:, :, :], duet_run.loading[1:, :, :]
        )
        assert math.isclose(
            np.mean(calibrated_duet.loading[0, :, :]), np.float32(0.6), abs_tol=10**-6
        )
        assert math.isclose(
            np.std(calibrated_duet.loading[0, :, :]), np.float32(0.3), abs_tol=10**-6
        )
        # try loading and depth
        grass_depth = assign_targets(method="meansd", mean=0.5, sd=0.05)
        loading_targets = set_fuel_parameter(parameter="loading", grass=grass_loading)
        depth_targets = set_fuel_parameter(parameter="depth", grass=grass_depth)
        calibrated_duet = calibrate(
            duet_run, fuel_parameter_targets=[loading_targets, depth_targets]
        )

        # try with two fueltypes
        litter_loading = assign_targets(method="meansd", mean=0.05, sd=0.005)
        loading_targets = set_fuel_parameter(
            parameter="loading", litter=litter_loading, grass=grass_loading
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=loading_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.loading, np.ndarray)
        assert np.allclose(calibrated_duet.depth, duet_run.depth)
        assert np.allclose(calibrated_duet.loading, duet_run.loading) == False

        # try with three fueltypes
        coniferous_loading = assign_targets(method="meansd", mean=0.1, sd=0.03)
        deciduous_loading = assign_targets(method="meansd", mean=0.5, sd=0.2)
        loading_targets = set_fuel_parameter(
            parameter="loading",
            coniferous=coniferous_loading,
            deciduous=deciduous_loading,
            grass=grass_loading,
        )
        # there is no deciduous litter, so it will throw an error
        with pytest.raises(ValueError):
            calibrated_duet = calibrate(
                duet_run, fuel_parameter_targets=loading_targets
            )
        # now without deciduous litter targets
        loading_targets = set_fuel_parameter(
            parameter="loading",
            coniferous=coniferous_loading,
            grass=grass_loading,
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=loading_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.loading, np.ndarray)
        assert np.allclose(calibrated_duet.depth, duet_run.depth)
        assert np.allclose(calibrated_duet.loading, duet_run.loading) == False
        assert math.isclose(
            np.mean(calibrated_duet.loading[0, :, :]), np.float32(0.6), abs_tol=10**-6
        )
        assert math.isclose(
            np.std(calibrated_duet.loading[0, :, :]), np.float32(0.3), abs_tol=10**-6
        )

    def test_constant_calibration(self):
        duet_run = import_duet(DATA_DIR / "v2")
        grass_depth = assign_targets(method="constant", value=0.5)
        litter_depth = assign_targets(method="constant", value=0.05)
        depth_targets = set_fuel_parameter(
            parameter="depth", grass=grass_depth, litter=litter_depth
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=depth_targets)
        assert isinstance(calibrated_duet.depth, np.ndarray)
        assert math.isclose(
            np.mean(calibrated_duet.depth[0, :, :][calibrated_duet.depth[0, :, :] > 0]),
            np.float32(0.5),
            abs_tol=10**-6,
        )
        assert math.isclose(
            np.mean(
                calibrated_duet.depth[1:, :, :][calibrated_duet.depth[1:, :, :] > 0]
            ),
            np.float32(0.05),
            abs_tol=10**-6,
        )

    def test_fueltype_all(self):
        duet_run = import_duet(DATA_DIR / "v2")
        loading = assign_targets(method="maxmin", max=2.0, min=0.5)
        grass_depth = assign_targets(method="constant", value=0.75)
        litter_depth = assign_targets(method="constant", value=0.15)
        loading_targets = set_fuel_parameter(parameter="loading", all=loading)
        depth_targets = set_fuel_parameter(
            parameter="depth", grass=grass_depth, litter=litter_depth
        )
        calibrated_duet = calibrate(
            duet_run,
            fuel_parameter_targets=[loading_targets, depth_targets],
        )
        assert isinstance(calibrated_duet, DuetRun)
        # first assert that the depth values are what we set them to be
        assert math.isclose(
            np.mean(calibrated_duet.depth[0, :, :][calibrated_duet.depth[0, :, :] > 0]),
            np.float32(0.75),
            abs_tol=10**-6,
        )
        assert math.isclose(
            np.mean(calibrated_duet.depth[1, :, :][calibrated_duet.depth[1, :, :] > 0]),
            np.float32(0.15),
            abs_tol=10**-6,
        )
        # then assert that a) the loading array has three layers, and b) the overall max and min matches
        assert calibrated_duet.loading.shape[0] == 3
        assert np.max(np.sum(calibrated_duet.loading, axis=0)) == np.float32(2.0)
        assert np.min(np.sum(calibrated_duet.loading, axis=0)) == np.float32(0.5)


def plot_array(x, title):
    plt.figure(2)
    plt.set_cmap("viridis")
    plt.imshow(x, origin="lower")
    plt.colorbar()
    plt.title(title, fontsize=18)
    plt.show()



from duet_tools.calibration import (
    DuetRun,
    Targets,
    FuelParameter,
    import_duet,
    import_duet_manual,
    assign_targets,
    set_loading,
    set_moisture,
    set_depth,
    set_fuel_parameter,
    calibrate,
)

from duet_tools.utils import (
    write_array_to_dat,
    read_dat_to_array,
)

from duet_tools.inputs import InputFile


__all__ = [
    "DuetRun",
    "Targets",
    "FuelParameter",
    "import_duet",
    "import_duet_manual",
    "assign_targets",
    "set_loading",
    "set_moisture",
    "set_depth",
    "set_fuel_parameter",
    "calibrate",
    "assign_targets_from_sb40",
    "LandfireQuery",
    "query_landfire",
    "write_array_to_dat",
    "read_dat_to_array",
    "InputFile",
]


def _missing_landfire(*args, **kwargs):
    raise ImportError(
        "The 'landfire' module requires additional dependencies. "
        "Please reinstall with: pip install package-name[landfire]"
    )


try:
    from duet_tools.landfire import (
        LandfireQuery,
        query_landfire,
        assign_targets_from_sb40,
    )

    __all__.extend(["LandfireQuery", "query_landfire", "assign_targets_from_sb40"])
except ImportError:
    LandfireQuery = _missing_landfire
    query_landfire = _missing_landfire
    assign_targets_from_sb40 = _missing_landfire



"""
DUET Tools Calibration module
"""

from __future__ import annotations

# Core Imports
from pathlib import Path
import warnings

# External Imports
import numpy as np
import pandas as pd

# Internal Imports
from duet_tools.utils import read_dat_to_array, write_array_to_dat
from duet_tools.inputs import InputFile

DATA_DIR = Path(__file__).parent / "data"

DATA_DIR = Path(__file__).parent / "data"


class DuetRun:
    """
    Class containing all arrays for a DUET run.

    Attributes
    ----------
    loading : np.ndarray
        3D Array of fuel loading (bulk density in kg/m^3) values in the format exported by DUET:
        Grass fuel load in first layer, litter fuel load for each tree species in
        subsequent layers.
    moisture : np.ndarray
        3D Array of fuel moisture content (%) values in the format exported by DUET:
        Grass moisture content in first layer, litter moisture content for each tree
        species in subsequent layers.
    depth : np.ndarray
        3D Array of fuelbd depth (m) values in the format exported by DUET:
        Grass height in first layer, litter depth for each tree species in subsequent
        layers.
    duet_version : str
        DUET version. Must be one of "v1" or "v2".

    """

    def __init__(
        self,
        loading: np.ndarray,
        depth: np.ndarray,
        moisture: np.ndarray,
        duet_version: str,
    ):
        self.loading = loading
        self.moisture = moisture
        self.depth = depth
        self.duet_version = duet_version

    def to_quicfire(
        self,
        directory: str | Path,
        loading: bool = True,
        moisture: bool = True,
        depth: bool = True,
        overwrite: bool = False,
    ) -> None:
        """
        Writes a DuetRun object to QUIC-fire fuel .dat inputs to a directory:
        treesrhof.dat, treesmoist.dat, treesfueldepth.dat

        Parameters
        ----------
        directory : str | Path
            Path to directory for writing QUIC-fire files
        loading : bool
            Whether to export the fuel loading array. Defaults to True.
        moisture : bool
            Whether to export the moisture content array. Defaults to True.
        depth : bool
            Whether to export the fuelbed depth array. Defaults to True.
        overwrite : bool
            Whether to overwrite trees*.dat files already present in the directory.
            If files exist, raises an error if True, warning if False. Defaults to False.

        Returns
        -------
        None :
            Writes QUIC-Fire .dat files to the provided directory.
        """
        written_files = []
        if isinstance(directory, str):
            directory = Path(directory)

        files = ["treesrhof.dat", "treesmoist.dat", "treesfueldepth.dat"]
        to_overwrite = []
        for file in files:
            path = directory / file
            if path.exists():
                to_overwrite.append(file)
        if len(to_overwrite) > 0:
            if overwrite:
                warnings.warn(
                    f"File(s) {to_overwrite} already exist(s) and will be overwritten."
                )
            else:
                raise FileExistsError(
                    f"File(s) {to_overwrite} already exist(s) in {directory}. "
                    f"Please set overwrite = True."
                )
        if loading:
            if self.loading is not None:
                treesrhof = self._integrate_all("loading")
                write_array_to_dat(treesrhof, "treesrhof.dat", directory)
                written_files.append("treesrhof.dat")
        if moisture:
            if self.moisture is not None:
                treesmoist = self._integrate_all("moisture")
                write_array_to_dat(treesmoist, "treesmoist.dat", directory)
                written_files.append("treesmoist.dat")
        if depth:
            if self.depth is not None:
                treesfueldepth = self._integrate_all("depth")
                write_array_to_dat(treesfueldepth, "treesfueldepth.dat", directory)
                written_files.append("treesfueldepth.dat")
        if len(written_files) == 0:
            print("No files were written")
        else:
            print(
                f"QUIC-Fire files {written_files} were written to directory {directory}"
            )

    def to_numpy(self, fuel_type: str, fuel_parameter: str) -> np.ndarray:
        """
        Returns a numpy array of the provided fuel type and parameter.

        Parameters
        ----------
        fuel_type : str
            Fuel type of desired array. Must be one of "integrated, "separated", "grass",
            "litter", "deciduous", or "coniferous.
            "integrated" : returns a vertically-integrated array of all fuel types.
                Array remains 3D, with shape (1,ny,nx). Integration method depends on
                fuel parameter.
            "separated" : returns a 3D array of shape (nlitter,ny,nx), where the first layer
                is grass, and the subsequent layers are litter. If using DUET v1, nlitter = 1;
                if using DUET v2, nlitter is 2 (deciduous and coniferous).
            "grass" : returns a 3D array of the chosen parameter for grass, with shape (1,ny,nx)
            "litter" : returns a 3D array of integrated litter values for all tree species,
            with shape (1,ny,nx).
            "coniferous" : returns a 3D array of litter values for coniferous tree species, with
            shape (1,ny,nx).
            "deciduous" : returns a 3D array of litter values for deciduous tree species, with
            shape (1,ny,nx).

        fuel_parameter : str
            Fuel parameter of desired array. Must be one of "loading", "moisture", or
            "depth".

        Returns
        -------
        np.ndarray :
            Numpy array of the provided fuel type and parameter.
        """
        self._validate_fuel_inputs(fuel_type, fuel_parameter)
        if fuel_type == "separated":
            return self.__dict__[fuel_parameter].copy()
        if fuel_type == "integrated":
            return self._integrate_all(fuel_parameter)
        if fuel_type == "grass":
            return self.__dict__[fuel_parameter][0, :, :].copy()
        if fuel_type == "litter":
            return self._integrate_litter(fuel_parameter)
        if fuel_type == "coniferous":
            return self.__dict__[fuel_parameter][1, :, :].copy()
        if fuel_type == "deciduous":
            return self.__dict__[fuel_parameter][2, :, :].copy()

    def _integrate_all(self, fuel_parameter: str) -> np.ndarray:
        if fuel_parameter == "loading":
            return np.sum(self.loading, axis=0)
        if fuel_parameter == "moisture":
            return _loading_weighted_average(self.moisture, self.loading)
        if fuel_parameter == "depth":
            return np.max(self.depth, axis=0)

    def _integrate_litter(self, fuel_parameter: str) -> np.ndarray:
        if fuel_parameter == "loading":
            return np.sum(self.loading[1:, :, :], axis=0)
        if fuel_parameter == "moisture":
            return _loading_weighted_average(
                self.moisture[1:, :, :], self.loading[1:, :, :]
            )
        if fuel_parameter == "depth":
            return np.max(self.depth[1:, :, :], axis=0)

    def _validate_input_moisture(self, moisture: np.ndarray):
        if moisture.shape != self.loading.shape:
            raise ValueError(
                f"Input array shape {moisture.shape} must match existing arrays {self.loading.shape}."
            )
        if self.loading[np.where(moisture == 0)].any() != 0:
            raise ValueError(
                "Value of moisture array cannot be zero where fuel is present"
            )

    def _validate_fuel_inputs(self, fuel_type: str, fuel_parameter: str):
        fueltypes_allowed = {
            "v1": ["grass", "litter", "separated", "integrated"],
            "v2": [
                "grass",
                "litter",
                "separated",
                "integrated",
                "deciduous",
                "coniferous",
            ],
        }
        if fuel_type not in fueltypes_allowed.get(self.duet_version):
            raise ValueError(
                f"Fuel type {fuel_type} not supported for DUET version {self.duet_version}. "
                f"Must be one of {fueltypes_allowed.get(self.duet_version)}"
            )
        parameters_allowed = ["loading", "moisture", "depth"]
        if fuel_parameter not in parameters_allowed:
            raise ValueError(
                f"Fuel parameter {fuel_parameter} not supported. Must be one of {parameters_allowed}"
            )


class Targets:
    """
    Class containing and validating target methods and values for fuel parameters.
    Should be instantiated using [`assign_targets`](reference.md#duet_tools.calibration.assign_targets).

    Attributes
    ----------
    method : str
        Method by which to calibrate to the target values. Must be one of
        "maxmin", "meansd", or "constant".
    args : list[str]
        Sting(s) to be used as keyword arguments for calibration, which correspond
        to the calibration method. For maxmin calibration, use ["max","min"];
        for meansd calibration, use ["mean","sd"]; for constant calibration, use
        []"value"].
    targets : list
        Calibration targets, which correspond to the elements of Targets.args.
    """

    def __init__(self, method: str, args: list[str], targets: list):
        self.method = self._validate_method(method)
        self.args, self.targets = self._validate_target_args(method, args, targets)
        self.calibration_function = self._get_calibration_function(method)

    def _get_calibration_function(self, method):
        if method == "maxmin":
            return _maxmin_calibration
        if method == "meansd":
            return _meansd_calibration
        if method == "constant":
            return _constant_calibration

    def _validate_method(self, method: str):
        methods_allowed = ["maxmin", "meansd", "constant"]
        if method not in methods_allowed:
            raise ValueError(
                f"Method {method} not supported. Must be one of {methods_allowed}"
            )
        return method

    def _validate_target_args(self, method: str, args: list[str], targets: list[float]):
        method_dict = {
            "constant": ["value"],
            "maxmin": ["max", "min"],
            "meansd": ["mean", "sd"],
        }
        args_allowed = method_dict.get(method)
        if set(args_allowed) != set(args):
            raise ValueError(f"Invalid **kwargs for method {method}. Must be {args}")

        targets_dict = dict(zip(args, targets))
        if method == "maxmin":
            if targets_dict["max"] <= targets_dict["min"]:
                raise ValueError("Target maximum must be greater than target minimum")
        if method == "meansd":
            if targets_dict["mean"] < targets_dict["sd"]:
                warnings.warn(
                    "Target mean is smaller than target sd. Were they input correctly?"
                )

        return args, targets


class FuelParameter:
    """
    Class containing and validating calibration targets for a single fuel parameter.
    A single Target object can be set for multiple fuel types. Should be instantiated using
    [`set_fuel_parameter`](reference.md#duet_tools.calibration.set_fuel_parameter)

    Attributes
    ----------
    parameter : str
        Fuel parameter for which targets should be set. Must be one of "loading", "moisture", or "depth".
    fuel_types : list[str]
        Fuel type(s) to which targets should be set. May be any of "grass", "litter",
        "coniferous", "deciduous", or "all".
    targets : list[Targets]
        Targets to be set to the provided parameter and fuel types.
    """

    def __init__(self, parameter: str, fuel_types: list[str], targets: list[Targets]):
        self.parameter = self._validate_fuel_parameter(parameter)
        self.fuel_types = self._validate_fuel_types(fuel_types)
        self.targets = targets

    def _validate_fuel_types(self, fuel_types):
        fueltypes_allowed = ["grass", "litter", "coniferous", "deciduous", "all"]
        for fuel_type in fuel_types:
            if fuel_type not in fueltypes_allowed:
                raise ValueError(
                    f"Method {fuel_type} not supported. Must be one of {fueltypes_allowed}"
                )
        if "all" in fuel_types and len(fuel_types) > 1:
            raise ValueError(
                "When fuel parameter targets are assigned to all fuel types, "
                "no other fuel parameter objects should be provided"
            )
        return fuel_types

    def _validate_fuel_parameter(self, parameter):
        fuel_parameters_allowed = ["loading", "moisture", "depth"]
        if parameter not in fuel_parameters_allowed:
            raise ValueError(
                f"Fuel parameter {parameter} not supported. Must be one of {fuel_parameters_allowed}"
            )
        return parameter


def import_duet_manual(
    directory: str | Path,
    loading_grid_name: str,
    moisture_grid_name: str,
    depth_grid_name: str,
    nx: int,
    ny: int,
    nsp: int,
    version: str,
) -> DuetRun:
    """
    Creates a DuetRun object from DUET output files

    Parameters
    ----------
    directory : str | Path
        Path to directory storing the DUET output files surface_rhof.dat and surface_depth.dat
    loading_grid_name: str
        File name of fuel loading (bulk density) DUET output.
    moisture_grid_name: str
        File name of fuel moisture DUET output.
    depth_grid_name: str
        File name of fuelbed depth DUET output.
    nx: int
        Number of DUET domain cells in the x-direction.
    ny: int
        Number of DUET domain cells in the y-direction.
    nsp: int
        Number of vegetation species (tree species + grass) in the DUET outputs.
        Must be 2 (grass and litter) for DUET v1.
    version: str
        DUET version that produced the outputs. Must be one of ["v1","v2"].

    Returns
    -------
    Instance of class DuetRun
    """
    supported = ["v1", "v2"]
    if version not in supported:
        raise ValueError(
            f"Version {version} not supported. Please use one of {supported}"
        )
    if isinstance(directory, str):
        directory = Path(directory)

    loading_nsp = read_dat_to_array(
        directory=directory,
        filename=loading_grid_name,
        nx=nx,
        ny=ny,
        nsp=nsp,
    )
    depth_nsp = read_dat_to_array(
        directory=directory,
        filename=depth_grid_name,
        nx=nx,
        ny=ny,
        nsp=nsp,
    )
    moisture_nsp = read_dat_to_array(
        directory=directory,
        filename=moisture_grid_name,
        nx=nx,
        ny=ny,
        nsp=nsp,
    )
    if version == "v1":  # arrays are kept as-is
        loading = loading_nsp.copy()
        depth = depth_nsp.copy()
        moisture = moisture_nsp.copy()
    if version == "v2":
        loading = np.zeros((3, ny, nx))
        depth = np.zeros((3, ny, nx))
        moisture = np.zeros((3, ny, nx))
        loading[0, :, :] = loading_nsp[0, :, :]
        depth[0, :, :] = depth_nsp[0, :, :]
        moisture[0, :, :] = moisture_nsp[0, :, :]

        # identify which species are deciduous and coniferous
        groups = _group_litter_species(directory)
        coniferous_indices = [i for i, v in groups.items() if v == "coniferous"]
        deciduous_indices = [i for i, v in groups.items() if v == "deciduous"]

        # for each parameter, coniferous is layer 1, deciduous is layer 2
        loading[1, :, :] = (
            loading_nsp[coniferous_indices].sum(axis=0)
            if coniferous_indices
            else np.zeros(loading.shape[1:])
        )
        loading[2, :, :] = (
            loading_nsp[deciduous_indices].sum(axis=0)
            if deciduous_indices
            else np.zeros(loading.shape[1:])
        )
        depth[1, :, :] = (
            depth_nsp[coniferous_indices].sum(axis=0)
            if coniferous_indices
            else np.zeros(depth.shape[1:])
        )
        depth[2, :, :] = (
            depth_nsp[deciduous_indices].sum(axis=0)
            if deciduous_indices
            else np.zeros(depth.shape[1:])
        )
        moisture[1, :, :] = (
            _loading_weighted_average(
                moisture_nsp[coniferous_indices], loading_nsp[coniferous_indices]
            )
            if coniferous_indices
            else np.zeros(moisture.shape[1:])
        )
        moisture[2, :, :] = (
            _loading_weighted_average(
                moisture_nsp[deciduous_indices], loading_nsp[deciduous_indices]
            )
            if deciduous_indices
            else np.zeros(loading.shape[1:])
        )

    return DuetRun(
        loading=loading, depth=depth, moisture=moisture, duet_version=version
    )


def import_duet(directory: Path | str, version: str = "v2") -> DuetRun:
    """
    Creates a DuetRun object from DUET input and output files. Assumes all
    files from a DUET run are present and unaltered. To import a DUET run
    manually, use `import_duet_manual`.

    Parameters
    ----------
    directory : str | Path
        Path to directory storing the DUET output files surface_rhof.dat and surface_depth.dat
        and the DUET input files duet.in and treesspcd.dat
    version: str
        DUET version that produced the outputs. Must be one of ["v1","v2"]. Defaults to "v2".

    Returns
    -------
    Instance of class DuetRun
    """
    if isinstance(directory, str):
        directory = Path(directory)

    supported = ["v1", "v2"]
    if version not in supported:
        raise ValueError(
            f"Version {version} not supported. Please use one of {supported}"
        )

    input_file = InputFile.from_directory(directory)
    nx = input_file.nx
    ny = input_file.ny
    species_list = _read_surface_species(directory)
    nsp = len(species_list) + 1  # number of tree species plus grass

    name_dict = {
        "rhof": {"v1": "surface_rhof.dat", "v2": "surface_rhof_layered.dat"},
        "depth": {"v1": "surface_depth.dat", "v2": "surface_depth_layered.dat"},
        "moist": {"v1": "surface_moist.dat", "v2": "surface_moist_layered.dat"},
    }

    loading_grid_name = name_dict["rhof"].get(version)
    moisture_grid_name = name_dict["moist"].get(version)
    depth_grid_name = name_dict["depth"].get(version)

    return import_duet_manual(
        directory,
        loading_grid_name,
        moisture_grid_name,
        depth_grid_name,
        nx,
        ny,
        nsp,
        version,
    )


def assign_targets(method: str, **kwargs: float) -> Targets:
    """
    Assigns target values and calculation method for exactly one fuel type and parameter

    Parameters
    ----------
    method : str
        Calibration method for the target values provided. Must be one of:
        "constant", "maxmin", "meansd", "sb40".
    **kwargs : float
        Keyword arguments correspond to the calibration method.
        For "maxmin" method, **kwargs keys must be `max` and `min`.
        For "meansd" method, **kwargs keys must be `mean` and `sd`.
        For "constant" method, **kwargs key must be `value`.

    Returns
    -------
    Instance of class Targets
    """
    args = list(kwargs.keys())
    targets = list(kwargs.values())

    return Targets(method=method, args=args, targets=targets)


def set_fuel_parameter(parameter: str, **kwargs: Targets):
    """
    Sets calibration targets for grass, litter, both separately, or all
    fuel types together, for a single fuel parameter.

    Parameters
    ----------
    parameter : str
        Fuel parameter for which to set targets
    **kwargs : Targets
        grass : Targets
            Grass calibration targets. Only the grass layer of the DUET parameter
            array will be calibrated.
        litter : Targets
            Litter calibration targets. Only the litter layer(s) of the DUET
            parameter array will be calibrated. Coniferous and deciduous litter will
            be calibrated together.
        coniferous : Targets
            Coniferous litter calibration targets. Only the coniferous litter layer of
            the DUET parameter array will be calibrated.
        deciduous : Targets
            Deciduous litter calibration targets. Only the deciduous litter layer of
            the DUET parameter array will be calibrated.
        all : Targets
            Calibration targets for all (both) fuel types. All layers of the
            DUET parameter array will be calibrated together.

    Returns
    -------
    FuelParameter :
        Object representing targets for the given fuel parameter, for each provided fuel type
    """
    parameter = parameter
    fuel_types = list(kwargs.keys())
    targets = list(kwargs.values())

    return FuelParameter(parameter, fuel_types, targets)


def set_loading(**kwargs: Targets):
    """
    Sets fuel loading calibration targets for grass, litter, both separately, or all
    fuel types together.

    Parameters
    ----------
    grass : Targets | None
        Grass calibration targets. Only the grass layer of the DUET fuel loading
        array will be calibrated.
    litter : Targets | None
        Litter calibration targets. Only the litter layer(s) of the DUET
        fuel loading array will be calibrated. Coniferous and deciduous litter will
        be calibrated together.
    coniferous : Targets | None
        Coniferous litter calibration targets. Only the coniferous litter layer of
        the DUET parameter array will be calibrated.
    deciduous : Targets | None
        Deciduous litter calibration targets. Only the deciduous litter layer of
        the DUET fuel loading array will be calibrated.
    all : Targets | None
        Calibration targets for all (both) fuel types. All layers of the
        DUET fuel loading array will be calibrated together.

    Returns
    -------
    FuelParameter :
        Object representing fuel loading targets for each provided fuel type
    """
    parameter = "loading"
    fuel_types = list(kwargs.keys())
    targets = list(kwargs.values())

    return FuelParameter(parameter, fuel_types, targets)


def set_moisture(**kwargs: Targets):
    """
    Sets moisture calibration targets for grass, litter, both separately, or all
    fuel types together.

    Parameters
    ----------
    grass : Targets | None
        Grass calibration targets. Only the grass layer of the DUET moisture
        array will be calibrated.
    litter : Targets | None
        Litter calibration targets. Only the litter layer(s) of the DUET
        moisture array will be calibrated. Coniferous and deciduous litter will
        be calibrated together.
    coniferous : Targets | None
        Coniferous litter calibration targets. Only the coniferous litter layer of
        the DUET parameter array will be calibrated.
    deciduous : Targets | None
        Deciduous litter calibration targets. Only the deciduous litter layer of
        the DUET moisture array will be calibrated.
    all : Targets | None
        Calibration targets for all (both) fuel types. All layers of the
        DUET moisture array will be calibrated together.

    Returns
    -------
    FuelParameter :
        Object representing moisture targets for each provided fuel type
    """
    parameter = "moisture"
    fuel_types = list(kwargs.keys())
    targets = list(kwargs.values())

    return FuelParameter(parameter, fuel_types, targets)


def set_depth(**kwargs: Targets):
    """
    Sets fuelbed depth calibration targets for grass, litter, both separately, or all
    fuel types together.

    Parameters
    ----------
    grass : Targets | None
        Grass calibration targets. Only the grass layer of the DUET depth
        array will be calibrated.
    litter : Targets | None
        Litter calibration targets. Only the litter layer(s) of the DUET
        depth array will be calibrated. Coniferous and deciduous litter will
        be calibrated together.
    coniferous : Targets | None
        Coniferous litter calibration targets. Only the coniferous litter layer of
        the DUET parameter array will be calibrated.
    deciduous : Targets | None
        Deciduous litter calibration targets. Only the deciduous litter layer of
        the DUET depth array will be calibrated.
    all : Targets | None
        Calibration targets for all (both) fuel types. All layers of the
        DUET depth array will be calibrated together.

    Returns
    -------
    FuelParameter :
        Object representing depth targets for each provided fuel type
    """
    parameter = "depth"
    fuel_types = list(kwargs.keys())
    targets = list(kwargs.values())

    return FuelParameter(parameter, fuel_types, targets)


def calibrate(
    duet_run: DuetRun, fuel_parameter_targets: list[FuelParameter] | FuelParameter
) -> DuetRun:
    """
    Calibrates the arrays in a DuetRun object using the provided targets and methods for one
    or more fuel types.

    Parameters
    ----------
    duet_run : DuetRun
        The DUET run to calibrate

    fuel_type_targets : FuelParameters | list(FuelParameters)
        FuelParameters object or list of FuelParameters objects for the fuel types
        to be calibrated.

    Returns
    -------
    Instance of class DuetRun with calibrated fuel arrays
    """
    if isinstance(fuel_parameter_targets, FuelParameter):
        fuel_parameter_targets = [fuel_parameter_targets]

    calibrated_duet = _duplicate_duet_run(duet_run)
    for fuelparameter in fuel_parameter_targets:
        fuelparam = fuelparameter.parameter
        for i in range(len(fuelparameter.fuel_types)):
            fueltype = fuelparameter.fuel_types[i]
            array_to_calibrate = _get_array_to_calibrate(duet_run, fueltype, fuelparam)
            if np.sum(array_to_calibrate) == 0:
                raise ValueError(f"No fuels present for fuel type {fueltype}")
            calibrated_array = _do_calibration(
                array_to_calibrate, fuelparameter.targets[i]
            )
            calibrated_duet = _add_calibrated_array(
                calibrated_duet, calibrated_array, fueltype, fuelparam
            )
    return calibrated_duet


def get_unit_from_fastfuels(zroot):
    """
    Creates a geojson bounding box of a fastfuels domain.

    Returns
    -------
    geojson
    """
    # TODO: write get_unit_from_fastfuels


def get_unit_from_shapefile(directory: str | Path):
    """
    Reads in a shapefile and returns a geojson bounding box.

    Returns
    -------
    geojson
    """
    # TODO: write get_unit_from_shapefile


def write_numpy_to_quicfire(array: np.ndarray, directory: str | Path, filename: str):
    """
    Writes a numpy array to a QUIC-Fire fuel input (.dat) in the chosen directory.

    Parameters
    ----------
    array : np.ndarray
        The numpy array to be written. Must be 3D.
    directory : str | Path
        The directory where the file will be written.
    filename : str
        The name of the file to be written. Must end in ".dat".

    Returns
    -------
    None
        File is written to disk.
    """
    if isinstance(directory, str):
        directory = Path(directory)
    write_array_to_dat(array=array, dat_name=filename, output_dir=directory)


def _get_array_to_calibrate(duet_run: DuetRun, fueltype: str, fuelparam: str):
    """
    Identifies and returns the layer of the duet outputs should be processed based
    on the fuel parameter and fuel type.
    """
    if fuelparam == "loading":
        if fueltype == "grass":
            return duet_run.loading[0, :, :].copy()
        if fueltype == "coniferous":
            return duet_run.loading[1, :, :].copy()
        if fueltype == "deciduous":
            return duet_run.loading[2, :, :].copy()
        if fueltype == "litter":
            return duet_run.loading[1:, :, :].sum(axis=0)
        else:
            return np.sum(duet_run.loading, axis=0)
    if fuelparam == "depth":
        if fueltype == "grass":
            return duet_run.depth[0, :, :].copy()
        if fueltype == "coniferous":
            return duet_run.depth[1, :, :].copy()
        if fueltype == "deciduous":
            return duet_run.depth[2, :, :].copy()
        if fueltype == "litter":
            return duet_run.depth[1:, :, :].sum(axis=0)
        else:
            return np.max(duet_run.depth, axis=0)
    if fuelparam == "moisture":
        if fueltype == "grass":
            return duet_run.moisture[0, :, :].copy()
        if fueltype == "coniferous":
            return duet_run.moisture[1, :, :].copy()
        if fueltype == "deciduous":
            return duet_run.moisture[2, :, :].copy()
        if fueltype == "litter":
            return _loading_weighted_average(
                duet_run.moisture[1:, :, :], duet_run.loading[1:, :, :]
            )
        else:
            loading_weights = duet_run.loading.copy()
            loading_weights[loading_weights == 0] = 0.01
            return np.average(duet_run.moisture, weights=loading_weights, axis=0)


def _duplicate_duet_run(duet_run: DuetRun) -> DuetRun:
    """
    Makes a copy of a DuetRun object.
    """
    new_loading = duet_run.loading.copy() if duet_run.loading is not None else None
    new_moisture = duet_run.moisture.copy() if duet_run.moisture is not None else None
    new_depth = duet_run.depth.copy() if duet_run.depth is not None else None

    new_duet = DuetRun(
        loading=new_loading,
        moisture=new_moisture,
        depth=new_depth,
        duet_version=duet_run.duet_version,
    )

    return new_duet


def _do_calibration(array: np.ndarray, target_obj: Targets):
    """
    Calibrates an array based on the method and values in a Targets object.
    """
    kwarg_dict = {}
    for i in range(len(target_obj.args)):
        kwarg_dict[target_obj.args[i]] = target_obj.targets[i]
    new_array = target_obj.calibration_function(array, **kwarg_dict)
    return new_array


def _maxmin_calibration(x: np.ndarray, **kwargs: float) -> np.ndarray:
    """
    Scales and shifts values in a numpy array based on an observed range. Does not assume
    data is normally distributed.
    """
    max_val = kwargs["max"]
    min_val = kwargs["min"]
    x1 = x[x > 0]
    if np.max(x1) == np.min(x1):
        raise ValueError(
            "maxmin calibration cannot be used when array has only one positive value. "
            "Please use 'constant' calibration method"
        )
    x2 = (x1 - np.min(x1)) / (np.max(x1) - np.min(x1))
    x3 = x2 * (max_val - min_val)
    x4 = x3 + min_val
    xnew = x.copy()
    xnew[np.where(x > 0)] = x4
    return xnew


def _meansd_calibration(x: np.ndarray, **kwargs: float) -> np.ndarray:
    """
    Scales and shifts values in a numpy array based on an observed mean and standard deviation.
    Assumes data is normally distributed.
    """
    mean_val = kwargs["mean"]
    sd_val = kwargs["sd"]
    x1 = x[x > 0]
    if np.max(x1) == np.min(x1):
        raise ValueError(
            "meansd calibration should not be used when array has only one positive value. "
            "Please use 'constant' calibration method"
        )
    x2 = mean_val + (x1 - np.mean(x1)) * (sd_val / np.std(x1))
    xnew = x.copy()
    xnew[np.where(x > 0)] = x2
    if np.min(xnew) < 0:
        xnew = _truncate_at_0(xnew)
    return xnew


# TODO: add option for fuel vs cell bulk density?
def _constant_calibration(x: np.ndarray, **kwargs: float) -> np.ndarray:
    """
    Conducts constant calibration by changing all nonzero values in a
    duet fuel parameter/type to a single given value.
    """
    value = kwargs["value"]
    arr = x.copy()
    arr[arr > 0] = value
    return arr


def _add_calibrated_array(
    duet_to_calibrate: DuetRun,
    calibrated_array: np.ndarray,
    fueltype: str,
    fuelparam: str,
) -> DuetRun:
    """
    Replaces or creates calibrated array(s) in a DuetRun object.
    """
    for param in ["loading", "moisture", "depth"]:
        if fuelparam == param:
            if fueltype == "grass":
                duet_to_calibrate.__dict__[param][0, :, :] = calibrated_array
            if fueltype == "coniferous":
                duet_to_calibrate.__dict__[param][1, :, :] = calibrated_array
            if fueltype == "deciduous":
                duet_to_calibrate.__dict__[param][2, :, :] = calibrated_array
            if fueltype == "litter":
                duet_to_calibrate.__dict__[param][1:, :, :] = _separate_2d_array(
                    calibrated_array, param, fueltype, duet_to_calibrate
                )
            if fueltype == "all":
                duet_to_calibrate.__dict__[param] = _separate_2d_array(
                    calibrated_array, param, fueltype, duet_to_calibrate
                )
    return duet_to_calibrate


def _separate_2d_array(
    calibrated: np.ndarray, param: str, fueltype: str, duet_run: DuetRun
) -> np.ndarray:
    """
    Separates a combined array into its component fuel types based on the
    fuel parameter.
    """
    if fueltype == "all":
        separated = np.zeros(duet_run.loading.shape)
        weights = np.zeros(duet_run.loading.shape)
    if fueltype == "litter":
        separated = np.zeros(duet_run.loading[1:, :, :].shape)
        weights = np.zeros(duet_run.loading[1:, :, :].shape)
    if param == "loading":
        for s in range(separated.shape[0]):
            loading = (
                duet_run.loading[1:, :, :] if fueltype == "litter" else duet_run.loading
            )
            loading_sum = np.sum(loading, axis=0)
            weights[s, :, :] = np.where(
                loading_sum != 0, loading[s, :, :] / loading_sum, 0
            )
            separated[s, :, :] = calibrated * weights[s, :, :]
    if param == "moisture":
        for s in range(duet_run.loading.shape[0]):
            separated[s, :, :][np.where(duet_run.moisture[s, :, :] == 0)] = 0
    if param == "depth":
        for s in range(separated.shape[0]):
            depth = duet_run.depth[1:, :, :] if fueltype == "litter" else duet_run.depth
            if fueltype == "all":
                depth_max = np.max(depth, axis=0)
                weights[s, :, :] = np.where(
                    depth_max != 0, depth[s, :, :] / depth_max, 0
                )
            if fueltype == "litter":
                depth_sum = np.sum(depth, axis=0)
                weights[s, :, :] = np.where(
                    depth_sum != 0, depth[s, :, :] / depth_sum, 0
                )
            separated[s, :, :] = calibrated * weights[s, :, :]
    return separated


def _truncate_at_0(arr: np.ndarray) -> np.ndarray:
    """
    Artificially truncates data to positive values by scaling all values below the median
    to the range (0, mean), effectively "compressing" those values.
    """
    arr2 = arr.copy()
    bottom_half = arr2[arr2 < np.median(arr2)]
    squeezed = (bottom_half - np.min(bottom_half)) / (
        np.max(bottom_half) - np.min(bottom_half)
    ) * (np.median(arr2) - 0) + 0
    arr2[np.where(arr2 < np.median(arr2))] = squeezed
    arr2[np.where(arr == 0)] = 0
    return arr2


def _loading_weighted_average(moisture: np.ndarray, loading: np.ndarray) -> np.ndarray:
    """
    Vertically integrate moisture by a weighted mean, where the weights come from cell bulk density
    """
    weights = _maxmin_calibration(loading, max=1.0, min=0)
    weights[weights == 0] = 0.01
    masked = np.ma.masked_array(moisture, moisture == 0)
    averaged = np.ma.average(masked, axis=0, weights=weights)
    integrated = np.ma.filled(averaged, 0)
    return integrated


# def _read_treesspcd(dir: Path, nx: int, ny: int, nz: int) -> int:
#     """
#     Interprets number of species from treesspcd.dat. Broken right now.
#     """
#     treesspcd = read_dat_to_array(dir, "treesspcd.dat", nx, ny, nz=nz, dtype=np.int32)
#     nsp = len(np.unique(treesspcd))
#     return nsp  # number of tree species + grass


def _read_surface_species(dir: Path) -> list:
    with open(dir / "surface_species.dat") as dat:
        species = dat.readlines()
    return species


def _group_litter_species(dir: Path) -> dict:
    """
    Returns a dictionary indicating whether the tree species is deciduous
    or coniferous for each layer in the surface_*_layered.dat files.
    """
    ref_species = pd.read_csv(DATA_DIR / "REF_SPECIES.csv")
    ref_species["Group"] = ref_species["MAJOR_SPGRPCD"].apply(_classify_spgrpcd)
    lookup_dict = (
        ref_species.drop_duplicates(subset="SPCD").set_index("SPCD")["Group"].to_dict()
    )

    with open(dir / "surface_species.dat", "r") as dat:
        lines = dat.readlines()

    spcd = [int(line.strip()) for line in lines]
    spcd_dict = {}
    for i in range(len(spcd)):
        if spcd[i] is None:
            raise Exception(
                f"DUET run contains species with FIA code {spcd[i]} "
                f"which does not have a valid major species group code"
            )
        spcd_dict[i + 1] = lookup_dict.get(spcd[i])

    return spcd_dict


def _classify_spgrpcd(code):
    if code in [1, 2]:
        return "coniferous"
    elif code in [3, 4]:
        return "deciduous"
    else:
        return None



"""
Test module for the inputs module of the duet_tools package.
"""

from __future__ import annotations

import pytest
from pathlib import Path

from duet_tools.inputs import InputFile

TEST_DIR = Path(__file__).parent
TMP_DIR = TEST_DIR / "tmp"
TMP_DIR.mkdir(exist_ok=True)
DATA_DIR = TEST_DIR / "test-data"


class TestInputFile:
    def test_create_input_file(self):
        input_file = InputFile.create(
            nx=234, ny=345, nz=30, duration=5, wind_direction=225
        )
        # test that data types are correct
        assert isinstance(input_file, InputFile)
        assert isinstance(input_file.nx, int)
        assert isinstance(input_file.dx, float)
        assert isinstance(input_file.duration, int)
        assert isinstance(input_file.random_seed, int)

        # test that input values are correct
        assert input_file.ny == 345
        assert input_file.dy == 2.0
        assert input_file.wind_variability == 359

    def test_write_input_file(self):
        verification = [
            "234  ! number of cells in x direction\n",
            "345  ! number of cells in y direction\n",
            "30  ! number of cells in z direction\n",
            "2.0  ! cell size in x direction (in meters)\n",
            "2.0  ! cell size in y direction (in meters)\n",
            "1.0  ! cell size in z direction (in meters)\n",
            "47  ! random number seed\n",
            "225  ! wind direction (in degrees)\n",
            "359.0  ! wind direction variability (in degrees)\n",
            "5  ! duration of simulation (in years)\n",
        ]

        input_file = InputFile.create(
            nx=234, ny=345, nz=30, duration=5, wind_direction=225, random_seed=47
        )
        input_file.to_file(TMP_DIR)

        with open(TMP_DIR / "duet.in", "r") as f:
            lines = f.readlines()

        for i in range(len(lines)):
            assert lines[i] == verification[i]



from __future__ import annotations

from pathlib import Path
from duet_tools import (
    InputFile,
    import_duet,
    assign_targets,
    set_fuel_parameter,
    calibrate,
)

duet_path = Path(__file__).parent.parent / "tests" / "tmp"

# Create DUET input file
input_file = InputFile.create(nx=100, ny=100, nz=30, duration=5, wind_direction=270)
input_file.to_file(duet_path)

## DUET IS RUN LOCALLY

# Import DUET outputs
duet_run = import_duet(directory=duet_path)

# Assign targets for each fuel type and fuel parameter
## method options: "maxmin", "meansd", "constant"
grass_loading = assign_targets(method="maxmin", max=1.0, min=0.1)
litter_loading = assign_targets(method="meansd", mean=0.6, sd=0.05)
grass_depth = assign_targets(method="constant", value=1.0)
coniferous_depth = assign_targets(method="constant", value=0.03)
deciduous_depth = assign_targets(method="constant", value=0.1)

# Bring together fuel types for each parameter
loading_targets = set_fuel_parameter(
    parameter="loading", grass=grass_loading, litter=litter_loading
)
depth_targets = set_fuel_parameter(
    parameter="depth",
    grass=grass_depth,
    coniferous=coniferous_depth,
    deciduous=deciduous_depth,
)

# Calibrate the DUET run
calibrated_duet = calibrate(
    duet_run=duet_run, fuel_parameter_targets=[loading_targets, depth_targets]
)

# Look at individual numpy arrays
calibrated_litter_loading = calibrated_duet.to_numpy(
    fuel_type="litter", fuel_parameter="loading"
)  # 2D array
calibrated_depth = calibrated_duet.to_numpy(
    fuel_type="separated", fuel_parameter="depth"
)  # 3D array, three layers
calibrated_loading = calibrated_duet.to_numpy(
    fuel_type="integrated", fuel_parameter="loading"
)  # 2D array

# Export to QUIC-Fire .dat files
qf_path = Path(__file__).parent.parent / "tests" / "tmp"
calibrated_duet.to_quicfire(directory=qf_path)



"""
Functions for interfacing with the LandFire API and processing the outputs
"""

# Core imports
from pathlib import Path
import importlib.resources
import re
import warnings
import zipfile

# External imports
import numpy as np
import pandas as pd
import geojson
import shapely
from pyproj import Transformer
import landfire
from landfire.geospatial import get_bbox_from_polygon
import rasterio as rio

# Internal imports
from duet_tools.calibration import Targets

try:  # Python 3.9+
    DATA_PATH = importlib.resources.files("duet_tools").joinpath("data")
except AttributeError:  # Python 3.6-3.8
    from pkg_resources import resource_filename

    DATA_PATH = resource_filename("duet_tools", "data")


class LandfireQuery:
    """
    Class containing the information from a LandFire query, to be passed to assign_targets()
    """

    def __init__(
        self,
        fuel_types: np.ndarray,
        loading: np.ndarray,
        moisture: np.ndarray,
        depth: np.ndarray,
    ):
        self.fuel_types = fuel_types
        self.loading = loading
        self.moisture = moisture
        self.depth = depth
        self._validate_arrays_shape()

    def _get_fueltype_indices(self, arr: np.ndarray, ft: int):
        ft_dict = {1: "grass", -1: "litter"}
        if ft not in arr.flatten():
            raise ValueError(f"Landfire query does not contain any {ft_dict[ft]} fuel.")
        return np.where(arr == ft)

    def _validate_arrays_shape(self):
        attributes = [
            getattr(self, attr)
            for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]
        first_shape = None
        for arr in attributes:
            if isinstance(arr, np.ndarray):
                if first_shape is None:
                    first_shape = arr.shape
                elif arr.shape != first_shape:
                    raise ValueError(
                        "All arrays in LandfireQuery must have the same shape."
                    )

    def _validate_get_targets(self, fuel_type, parameter, method):
        fueltypes_allowed = ["grass", "litter", "all"]
        parameters_alowed = ["loading", "moisture", "depth"]
        methods_allowed = ["maxmin", "meansd", "constant"]

        if fuel_type not in fueltypes_allowed:
            raise ValueError(
                f"Fuel type {fuel_type} not supported. Must be one of {fueltypes_allowed}."
            )
        if parameter not in parameters_alowed:
            raise ValueError(
                f"Parameter {parameter} not supported. Must be one of {parameters_alowed}."
            )
        if method not in methods_allowed:
            raise ValueError(
                f"Method {method} not supported. Must be one of {methods_allowed}."
            )


def query_landfire(
    area_of_interest: geojson.Polygon | shapely.Polygon,
    year: str,
    directory: str | Path,
    input_epsg: int,
    delete_files: bool = True,
) -> LandfireQuery:
    """
    Creates and submits a LANDFIRE query for a specified area of interest.

    Parameters
    ----------
    area_of_interest : geojson.Polygon | shapely.Polygon
        Area in which to query LANDFIRE data. For best results, dimensions in meters should
        match (nx*dx, ny*dy) of DUET domain.
    year : int
        Year of LANDFIRE data to query. Must be one of [2019, 2020, 2022].
    directory : Path | str
        Directory where files associated with the LANDFIRE query will be saved.
    input_epsg : int
        EPSG number for CRS of area_of_interest polygon
    delete_files : bool = True
        Whether to delete intermediate files created in the process of querying LANDFIRE data. Defaults to True

    Returns
    -------
    LandfireQuery
    """
    if isinstance(directory, str):
        directory = Path(directory)

    if isinstance(area_of_interest, geojson.Polygon):
        area_of_interest = shapely.Polygon(area_of_interest["coordinates"][0])

    valid_years = [2019, 2020, 2022]
    if year not in valid_years:
        raise ValueError(
            f"SB40 data for year {year} not available. Must be one of {valid_years}"
        )

    if input_epsg != 4236:
        area_of_interest = _reproject_polygon(
            area_of_interest, input_epsg, target_epsg=4326
        )

    _query_landfire(poly=area_of_interest, year=year, directory=directory)
    landfire_arr = _landfire_to_array(directory)

    # Import SB40 FBFM parameters table
    sb40_params_path = DATA_PATH / "sb40_parameters.csv"
    sb40_params = pd.read_csv(sb40_params_path)

    # Generate dict of fastfuels bulk density values and apply to Landfire query
    sb40_dict = _get_sb40_fuel_params(sb40_params)
    sb40_arr = _get_sb40_arrays(landfire_arr, sb40_dict)

    if delete_files:
        _delete_intermediate_files(directory)

    return LandfireQuery(
        fuel_types=sb40_arr[0, :, :],
        loading=sb40_arr[1, :, :],
        moisture=sb40_arr[2, :, :],
        depth=sb40_arr[3, :, :],
    )


def assign_targets_from_sb40(
    query: LandfireQuery, fuel_type: str, parameter: str, method: str = "maxmin"
) -> Targets:
    """
    Assign a calibration target and method for a given fuel type and parameter.

    Parameters
    ----------
    query : LandfireQuery
        An object of class LandfireQuery created with query_landfire. Calibration targets
        will be calculated from these values.
    fuel_type : str
        The fuel type to obtain target values for. Must be one of "grass", "litter", or "all".
    parameter : str
        The fuel parameter to obtain target values for. Must be one of "loading", "moisture", or "depth".
    method : str
        The desired calibration method for the sb40-derived targets. Must be one of "maxmin", "meandsd",
        or "constant". Default is "maxmin". "constant" is only recommended if only one parameter value
        is present for the given fuel type. "meansd" is not recommended since values often do not follow a
        normal distribution.

    Returns
    -------
    Targets :
        A Targets object with values derived from Landfire and SB40 fuel models
    """
    query._validate_get_targets(fuel_type, parameter, method)
    # select fuel parameter
    if parameter == "loading":
        param_arr = query.loading
    elif parameter == "moisture":
        param_arr = query.moisture
    else:
        param_arr = query.depth
    # select fuel type
    if fuel_type == "grass":
        fuel_arr = param_arr[query._get_fueltype_indices(query.fuel_types, 1)]
    elif fuel_type == "litter":
        fuel_arr = param_arr[query._get_fueltype_indices(query.fuel_types, -1)]
    else:
        fuel_arr = param_arr
    # get targets based on method
    fuel_arr = fuel_arr[np.where(fuel_arr > 0)]
    if method == "maxmin":
        if np.max(fuel_arr) == np.min(fuel_arr):
            warnings.warn(
                f"There is only one value for {fuel_type} {parameter}. "
                "Setting calibration method to 'constant'",
                UserWarning,
            )
            method = "constant"
            args = ["value"]
            targets = [np.mean(fuel_arr)]
        else:
            method = "maxmin"
            args = ["max", "min"]
            targets = [np.max(fuel_arr), np.min(fuel_arr)]
        return Targets(
            method=method,
            args=args,
            targets=targets,
        )
    if method == "meansd":
        if np.max(fuel_arr) == np.min(fuel_arr):
            warnings.warn(
                f"There is only one value for {fuel_type} {parameter}. "
                "Setting calibration method to 'constant'",
                UserWarning,
            )
            method = "constant"
            args = ["value"]
            targets = [np.mean(fuel_arr)]
        else:
            method = "meansd"
            args = ["mean", "sd"]
            targets = [np.mean(fuel_arr), np.std(fuel_arr)]
        return Targets(
            method=method,
            args=args,
            targets=targets,
        )
    if method == "constant":
        if np.max(fuel_arr) != np.min(fuel_arr):
            raise ValueError(
                "Multiple values present in Landfire query. Please use either maxmin "
                "or meansd calibration method."
            )
        return Targets(method="constant", args=["value"], targets=[np.mean(fuel_arr)])


def _query_landfire(
    poly: shapely.Polygon,
    year: int,
    directory: Path,
) -> None:
    """
    Download a grid of SB40 fuel models from Landfire for the unit and convert to a numpy array
    """

    if year == 2019:
        layer = ["200F40_19"]
    if year == 2020:
        layer = ["200F40_20"]
    if year == 2022:
        layer = ["220F40_22"]

    bbox = get_bbox_from_polygon(aoi_polygon=poly, crs=4326)

    # Download Landfire data to output directory
    lf = landfire.Landfire(bbox, output_crs="5070")
    lf.request_data(layers=layer, output_path=Path(directory, "landfire_sb40.zip"))

    # Exctract tif from compressed download folder and rename
    with zipfile.ZipFile(Path(directory, "landfire_sb40.zip")) as zf:
        extension = ".tif"
        rename = "landfire_sb40.tif"
        info = zf.infolist()
        for file in info:
            if file.filename.endswith(extension):
                file.filename = rename
                zf.extract(file, directory)


def _landfire_to_array(
    directory: Path,
) -> np.ndarray:
    # Upsample landfire raster to the quicfire resolution
    with rio.open(Path(directory, "landfire_sb40.tif")) as sb:
        arr = sb.read(1)

    return arr


def _get_sb40_fuel_params(params: pd.DataFrame) -> dict:
    """
    Builds a dictionary of SB40 fuel parameter values and converts them to
    the official FastFuels units

    Returns:
        dict: SB40 parameters for each fuel model
    """

    # Convert tons/ac-ft to kg/m^3
    params["1_hr_kg_per_m3"] = params["1_hr_t_per_ac"] * 0.22417
    params["10_hr_kg_per_m3"] = params["10_hr_t_per_ac"] * 0.22417
    params["100_hr_kg_per_m3"] = params["100_hr_t_per_ac"] * 0.22417
    params["live_herb_kg_per_m3"] = params["live_herb_t_per_ac"] * 0.22417
    params["live_woody_kg_per_m3"] = params["live_woody_t_per_ac"] * 0.22417

    # Convert inverse feet to meters
    params["dead_1_hr_sav_ratio_1_per_m"] = (
        params["dead_1_hr_sav_ratio_1_per_ft"] * 3.2808
    )
    params["live_herb_sav_ratio_1_per_m"] = (
        params["live_herb_sav_ratio_1_per_ft"] * 3.2808
    )
    params["live_wood_sav_ratio_1_per_m"] = (
        params["live_wood_sav_ratio_1_per_ft"] * 3.2808
    )

    # Convert percent to ratio
    params["dead_fuel_extinction_moisture"] /= 100

    # Convert feet to meters
    params["fuel_bed_depth_m"] = params["fuel_bed_depth_ft"] * 0.3048

    # Compute wet loading
    params["wet_load"] = params["1_hr_kg_per_m3"] + params["live_herb_kg_per_m3"]

    # Compute a live herb curing factor alpha as a function of wet loading.
    # This is kind of a B.S. approach raised by Rod on a phone call with
    # Anthony on 02/28/2023. I don't like this at all, but it is a temporary
    # Fix for the BP3D team to run some simulations.
    # low_load_fuel_models = [
    params["alpha"] = [0.5 if rho > 1 else 1.0 for rho in params["wet_load"]]

    # Compute dry loading
    params["dry_herb_load"] = params["live_herb_kg_per_m3"] * params["alpha"]
    params["dry_load"] = params["1_hr_kg_per_m3"] + params["dry_herb_load"]

    # Compute SAV
    params["sav_1hr_ratio"] = params["1_hr_kg_per_m3"] / params["dry_load"]
    params["sav_1hr"] = params["sav_1hr_ratio"] * params["dead_1_hr_sav_ratio_1_per_m"]
    params["sav_herb_ratio"] = params["dry_herb_load"] / params["dry_load"]
    params["sav_herb"] = (
        params["sav_herb_ratio"] * params["live_herb_sav_ratio_1_per_m"]
    )
    params["sav"] = params["sav_1hr"] + params["sav_herb"]

    # Convert nan to 0
    params.fillna(0, inplace=True)

    # Create dictionary for assigning fuel types for DUET calibration
    duet_dict = {
        "NB": 0,  # 0 = NEUTRAL, i.e. not predominantly grass or litter
        "GR": 1,  # 1 = GRASS predominantly
        "GS": 1,
        "SH": 1,  # I am considering shrubs as grass
        "TU": 0,
        "TL": -1,  # -1 = LITTER predominantly
        "SB": 0,
    }

    # Add column to df with DUET designations
    pattern = r"[0-9]"  # take out numbers from fbfm_type strings
    params["fbfm_cat"] = params["fbfm_code"].apply(lambda x: re.sub(pattern, "", x))
    params["duet_fuel_type"] = params["fbfm_cat"].apply(lambda x: duet_dict.get(x))

    # Build the dictionary with fuel parameters for the Scott and Burgan 40
    # fire behavior fuel models. Dict format: key ->
    # [name, loading (tons/ac), SAV (1/ft), ext. MC (percent), bed depth (ft)]
    # Note: Eventually we want to get rid of this and just use the dataframe.
    # This is legacy from the old parameter table json.
    sb40_dict = {}
    for key in params["key"]:
        row = params[params["key"] == key]
        sb40_dict[key] = [
            row["fbfm_code"].values[0],
            row["dry_load"].values[0],
            row["sav"].values[0],
            row["dead_fuel_extinction_moisture"].values[0],
            row["fuel_bed_depth_m"].values[0],
            row["duet_fuel_type"].values[0],
        ]
    sb40_dict[-9999] = ["NA", 0.0, 0.0, 0.0, 0.0, 0]

    return sb40_dict


def _get_sb40_arrays(sb40_keys: np.ndarray, sb40_dict: dict) -> np.ndarray:
    """
    Use a dictionary of fuel loading and fuel types that correspond to SB40
    fuel models to assign those values across the study area.

    Fuel types are as follows:
    - 1: Predominantly grass. All cells with a GR, GS, or SH designation from SB40.
    - -1: Predominantly tree litter. All cells with a TL designation from SB40.
    - 0: Neither predominantly grass or tree litter. All other SB40 designations.

    Returns:
    3D np.ndarray:
    4 layers:
        1. fuel types
        2. fuel loading (bulk density) values as calculated by fastfuels
        3. fuel moisture content values as calculated by fastfuels
        4. fuelbed depth values as caluclated by fastfuels
    """
    val_idx = [5, 1, 3, 4]
    fuel_arr = np.zeros((4, sb40_keys.shape[0], sb40_keys.shape[1]))
    for i in range(len(val_idx)):
        layer_dict = {key: val[val_idx[i]] for key, val in sb40_dict.items()}
        layer = np.vectorize(layer_dict.get)(sb40_keys)
        fuel_arr[i, :, :] = layer

    return fuel_arr


def _reproject_polygon(poly: shapely.Polygon, input_epsg: int, target_epsg: int):
    coords = list(poly.exterior.coords)
    transformer = Transformer.from_crs(input_epsg, target_epsg, always_xy=True)
    transformed_coords = transformer.transform(*zip(*coords))
    transformed_poly = shapely.Polygon(list(zip(*transformed_coords)))

    return transformed_poly


def _delete_intermediate_files(directory: Path):
    # Name intermediate files
    temp = [
        "landfire_sb40.zip",
        "landfire_sb40.tif",
        "sb40_upsampled.tif",
        "sb40_cropped.tif",
    ]
    [Path(directory, file).unlink() for file in temp if Path(directory, file).exists()]



"""
Test module for the utils module of the duet_tools package.
"""

from __future__ import annotations

import pytest
import geojson
import numpy as np
from pathlib import Path
from scipy.io import FortranFile

from duet_tools.utils import (
    read_dat_to_array,
    write_array_to_dat,
    read_shapefile_to_geojson,
)

TEST_DIR = Path(__file__).parent
TMP_DIR = TEST_DIR / "tmp"
TMP_DIR.mkdir(exist_ok=True)
DATA_DIR = TEST_DIR / "test-data"


def get_test_array(dir: Path, dat: str, shape: tuple) -> np.ndarray:
    with open(Path(dir, dat), "rb") as fin:
        array = FortranFile(fin).read_reals(dtype=np.float32).reshape(shape, order="F")
    return array


class TestUtilFunctions:
    def test_read_dat(self):
        dir = DATA_DIR / "v2"
        dat = "treesrhof.dat"
        arr = read_dat_to_array(dir, dat, nx=333, ny=295, nz=84)
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (84, 295, 333)
        test = get_test_array(dir, dat, (333, 295, 84))
        assert arr.all() == test.T.all()
        # try with invalid shape
        with pytest.raises(ValueError):
            arr = read_dat_to_array(dir, dat, nx=333, ny=294, nz=84)
        # try with invalid order
        with pytest.raises(ValueError):
            arr = read_dat_to_array(dir, dat, nx=333, ny=295, nz=84, order="P")
        # try with nsp instead of nz
        dat = "surface_rhof_layered.dat"
        arr = read_dat_to_array(dir, dat, nx=333, ny=295, nsp=9)
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (9, 295, 333)
        test = get_test_array(dir, dat, (9, 333, 295))
        assert arr.all() == np.moveaxis(test, 1, 2).all()

    def test_write_dat(self):
        dir_in = DATA_DIR / "v2"
        dir_out = TMP_DIR
        dat_in = "treesrhof.dat"
        dat_out = "temp.dat"
        arr1 = get_test_array(dir_in, dat_in, (333, 295, 84)).T
        write_array_to_dat(arr1, dat_out, dir_out)
        arr2 = get_test_array(dir_out, dat_out, (333, 295, 84)).T
        assert arr1.all() == arr2.all()

    def test_shapefile_to_geojson(self):
        # test with .shp
        json = read_shapefile_to_geojson(DATA_DIR / "new_mixedcon_area.shp")
        assert isinstance(json, geojson.Polygon)
        # test with .zip
        json = read_shapefile_to_geojson(DATA_DIR / "new_mixedcon_area.zip")
        assert isinstance(json, geojson.Polygon)



"""
Utility functions for DUET tools modules
"""

from __future__ import annotations

import geojson
import numpy as np
from pathlib import Path
from scipy.io import FortranFile
import shapefile  # pyshp
import io
import zipfile


def read_dat_to_array(
    directory: str | Path,
    filename: str,
    nx: int,
    ny: int,
    nz: int = None,
    nsp: int = None,
    order: str = "F",
    dtype: type = np.float32,
) -> np.ndarray:
    """
    Reads a fortran binary file (.dat) to a numpy array

    Parameters
    ----------
    directory: Path | str
        Path to directory of the .dat file.
    filename: str
        Name of the .dat file
    nx : int
        Number of cells in the x-direction
    ny : int
        Number of cells in the y-direction
    nz : int
        Number of cells in the z-direction
    nsp: int
        Number of species
    order : str
        Order of the .dat file. Must be one of "C" or "F". Defaults to "F".
    dtype : type
        Data type of the array. Defaults to np.float32

    Returns
    -------
        A numpy array with shape (nz, ny, nx).
    """
    if order not in ["C", "F"]:
        raise ValueError('Order must be either "C" or "F".')
    if isinstance(directory, str):
        directory = Path(directory)

    if (nz is None) and (nsp is None):
        shape = (nx, ny)
    elif nz is None:
        shape = (nsp, nx, ny)
    elif nsp is None:
        shape = (nx, ny, nz)
    else:
        shape = (nsp, nx, ny, nz)

    with open(Path(directory, filename), "rb") as fin:
        array = FortranFile(fin).read_reals(dtype=dtype).reshape(shape, order=order)

    if (nz is None) and (nsp is None):
        return np.moveaxis(array, 1, 0)
    elif nz is None:
        return np.moveaxis(array, 2, 1)
    elif nsp is None:
        return np.transpose(array)
    else:
        return np.moveaxis(np.moveaxis(array, 3, 1), 3, 1)


def write_array_to_dat(
    array: np.ndarray,
    dat_name: str,
    output_dir: Path | str,
    dtype: type = np.float32,
    reshape: bool = False,
) -> None:
    """
    Write a numpy array to a fortran binary file (.dat).

    Parameters
    ----------
    array : np.ndarray
        numpy array to be written to a file
    dat_name : str
        Filename ending with .dat
    output_dir : Path | str
        Directory where file will be written
    dtype : type
        Data type of the array. Defaults to np.float32
    reshape: bool
        Whether to reshape the array. Array dimensions in duet-tools are either (nz,ny,nx)
        or (nsp,ny,nx) and will be written in row-major order, meaning that with column-major
        order (default for fortran), they will be (nx, ny, nz). Reshaping (nsp, ny, nx) arrays
        to (ny, nx, nsp) will result in the column-major order of (nsp, nx, ny), which is
        expected by DUET and LANL Trees. If True, reshaping will only be applied to 3D arrays.
        Defaults to False.
    """
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    # Reshape array from (nsp, ny, nx) to (ny, nx, nsp)
    if reshape:
        if len(array.shape) == 3:
            array = np.moveaxis(array, 0, 2).astype(dtype)
        else:
            array = array.astype(dtype)
    else:
        array = array.astype(dtype)

    # Written in row-major order
    with FortranFile(Path(output_dir, dat_name), "w") as f:
        f.write_record(array)


def read_shapefile_to_geojson(path: Path) -> geojson.Polygon:
    """
    Read a shapefile and convert to a geojson polygon. May be used to
    query LANDFIRE data. Assumes the shapefile has Polygon geometry.
    Only the first feature is converted to a geojson.

    Parameters
    ----------
    path : str
        Path to shapefile. File may be compressed (.zip), or uncompressed (.shp)
        with constituent files in the same directory.

    Returns
    -------
    A geojson Polygon object.
    """
    if path.suffix == ".zip":
        with zipfile.ZipFile(path, "r") as zip_ref:
            # Extract shapefile components to memory
            shp_name = next(
                (name for name in zip_ref.namelist() if name.endswith(".shp")), None
            )
            shx_name = next(
                (name for name in zip_ref.namelist() if name.endswith(".shx")), None
            )
            dbf_name = next(
                (name for name in zip_ref.namelist() if name.endswith(".dbf")), None
            )

            if not (shp_name and shx_name and dbf_name):
                raise FileNotFoundError("Zip must contain .shp, .shx, and .dbf files.")

            # Read each file into BytesIO
            shp_io = io.BytesIO(zip_ref.read(shp_name))
            shx_io = io.BytesIO(zip_ref.read(shx_name))
            dbf_io = io.BytesIO(zip_ref.read(dbf_name))

            reader = shapefile.Reader(shp=shp_io, shx=shx_io, dbf=dbf_io)
    else:
        if not path.exists():
            raise FileNotFoundError(f"Shapefile not found: {path}")
        reader = shapefile.Reader(str(path))

    shape = reader.shape(0)
    if shape.shapeType != shapefile.POLYGON:
        raise ValueError("Shapefile does not contain a polygon.")

    coords = shape.points
    parts = list(shape.parts) + [len(coords)]
    rings = [coords[parts[i] : parts[i + 1]] for i in range(len(parts) - 1)]
    polygon = geojson.Polygon([ring for ring in rings])

    reader.close()
    return polygon



import requests
from setuptools import find_packages, setup
import sys

installing_landfire = any(arg.startswith("duet-tools[landfire]") for arg in sys.argv)
using_python_312 = sys.version_info >= (3, 12)
if using_python_312 and installing_landfire:
    sys.exit(
        "Error: 'duet-tools[landfire]' is only available for Python <3.12.\n"
        "Use Python 3.10 or 3.11 to install with the 'landfire' extra."
    )


def read_file(fname):
    with open(fname, encoding="utf-8") as fd:
        return fd.read()


def get_version():
    """Get the most recent tag (including pre-releases)."""
    import requests

    url = "https://api.github.com/repos/nmc-cafes/duet-tools/releases"
    response = requests.get(url)
    response.raise_for_status()
    releases = response.json()

    if not releases:
        raise ValueError("No releases found on GitHub.")

    # Use the most recently published release (pre or not)
    releases.sort(key=lambda r: r["created_at"], reverse=True)
    version = releases[0]["tag_name"]
    return version[1:]  # Remove the leading "v" from the version number


def get_requirements(file_name: str):
    """Get the requirements from the requirements.txt file."""
    requirements = []
    with open(f"requirements/{file_name}.txt", encoding="utf-8") as fd:
        for line in fd:
            line = line.strip()
            if line and not line.startswith("#"):
                requirements.append(line)
    return requirements


NAME = "duet-tools"
DESCRIPTION = "Output management tools for LANL DUET program"
LONG_DESCRIPTION = read_file("README.md")
VERSION = get_version()
LICENSE = "MIT"
URL = "https://github.com/nmc-cafes/duet-tools"
PROJECT_URLS = {"Bug Tracker": f"{URL}/issues"}
INSTALL_REQUIRES = get_requirements("requirements")
EXTRAS_REQUIRE = {"landfire": get_requirements("requirements_landfire")}

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license=LICENSE,
    url=URL,
    project_urls=PROJECT_URLS,
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
    ],
    package_dir={"": "."},
    packages=find_packages(exclude=["docs", "tests"]),
    package_data={"duet_tools": ["data/*"]},
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=">=3.10",
)



"""
DUET inputs module
"""

from __future__ import annotations

# Core Imports
from pathlib import Path
from random import randint


class InputFile:
    """
    Class representing a DUET input file.

    Attributes
    ----------
    nx: int
        Number of cells in the x-direction.
    ny: int
        Number of cells in the y-direction.
    nz: int
        Number of cells in the z-direction.
    dx: float
        Cell size in the x-direction (m)
    dy: float
        Cell size in the y-direction (m)
    dz: float
        Cell size in the z-direction (m)
    random_seed: int
        Random number seed.
    wind_direction: float
        Wind direction (degrees).
    wind_variability: float
        Wind direction variability (degrees).
    duration:
        Duration of simulation (years).
    """

    def __init__(
        self,
        nx: int,
        ny: int,
        nz: int,
        dx: float,
        dy: float,
        dz: float,
        random_seed: int,
        wind_direction: float,
        wind_variability: float,
        duration: int,
    ):
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.dx = dx
        self.dy = dy
        self.dz = dz
        self.random_seed = random_seed
        self.wind_direction = wind_direction
        self.wind_variability = wind_variability
        self.duration = duration

    @classmethod
    def create(
        cls,
        nx: int,
        ny: int,
        nz: int,
        duration: int,
        wind_direction: float,
        dx: float = 2.0,
        dy: float = 2.0,
        dz: float = 1.0,
        random_seed: int = None,
        wind_variability: float = 359.0,
    ):
        if random_seed is None:
            random_seed = randint(1, 50000)
        return cls(
            nx,
            ny,
            nz,
            dx,
            dy,
            dz,
            random_seed,
            wind_direction,
            wind_variability,
            duration,
        )

    def to_file(self, directory: Path | str):
        """
        Writes a DUET input file to the specified path

        Parameters
        ----------
        directory: Path | str
            Directory for writing DUET input file

        Returns
        -------
        None:
            Writes duet.in to directory
        """
        if isinstance(directory, str):
            directory = Path(directory)

        out_path = directory / "duet.in"
        with open(out_path, "w") as f:
            f.write(f"{self.nx}  ! number of cells in x direction\n")
            f.write(f"{self.ny}  ! number of cells in y direction\n")
            f.write(f"{self.nz}  ! number of cells in z direction\n")
            f.write(f"{self.dx}  ! cell size in x direction (in meters)\n")
            f.write(f"{self.dy}  ! cell size in y direction (in meters)\n")
            f.write(f"{self.dz}  ! cell size in z direction (in meters)\n")
            f.write(f"{self.random_seed}  ! random number seed\n")
            f.write(f"{self.wind_direction}  ! wind direction (in degrees)\n")
            f.write(
                f"{self.wind_variability}  ! wind direction variability (in degrees)\n"
            )
            f.write(f"{self.duration}  ! duration of simulation (in years)\n")

    @classmethod
    def from_directory(cls, dir: Path | str):
        """
        Creates an instance of class InputFile from a directory with a DUET input deck

        Parameters
        ----------

        dir: Path | str
            Path to the directory containing the DUET input deck.

        Returns
        -------
        InputFile
        """
        if isinstance(dir, str):
            dir = Path(dir)

        with open(dir / "duet.in") as f:
            lines = f.readlines()

        return cls(
            nx=int(lines[0].strip().split("!")[0]),
            ny=int(lines[1].strip().split("!")[0]),
            nz=int(lines[2].strip().split("!")[0]),
            dx=float(lines[3].strip().split("!")[0]),
            dy=float(lines[4].strip().split("!")[0]),
            dz=float(lines[5].strip().split("!")[0]),
            random_seed=int(lines[6].strip().split("!")[0]),
            wind_direction=float(lines[7].strip().split("!")[0]),
            wind_variability=float(lines[8].strip().split("!")[0]),
            duration=int(lines[9].strip().split("!")[0]),
        )



"""
Test module for the landfire module of the duet_tools package.
"""

from __future__ import annotations

import pytest
import numpy as np
from pathlib import Path
import matplotlib.pyplot as plt
import geojson

from duet_tools.calibration import Targets
from duet_tools.landfire import LandfireQuery, query_landfire, assign_targets_from_sb40

TEST_DIR = Path(__file__).parent
TMP_DIR = TEST_DIR / "tmp"
TMP_DIR.mkdir(exist_ok=True)
DATA_DIR = TEST_DIR / "test-data"


class TestLandfireTargets:
    @classmethod
    def get_geojson(self):
        with open(DATA_DIR / "anderson_butte.geojson") as fid:
            sample_geojson = geojson.load(fid)
        return geojson.Polygon(sample_geojson["features"][0]["geometry"]["coordinates"])

    def test_query_landfire(self):
        sample_aoi = self.get_geojson()
        query = query_landfire(
            area_of_interest=sample_aoi, year=2019, directory=TMP_DIR, input_epsg=4326
        )
        assert isinstance(query, LandfireQuery)
        assert isinstance(query.fuel_types, np.ndarray)
        assert isinstance(query.loading, np.ndarray)
        assert isinstance(query.moisture, np.ndarray)
        assert isinstance(query.depth, np.ndarray)

    def test_query_landfire_years(self):
        sample_aoi = self.get_geojson()
        # Query for 2019 is above. Test 2020...
        query = query_landfire(
            area_of_interest=sample_aoi, year=2020, directory=TMP_DIR, input_epsg=4326
        )
        # ...and 2022
        query = query_landfire(
            area_of_interest=sample_aoi, year=2022, directory=TMP_DIR, input_epsg=4326
        )
        # make sure no other years work
        with pytest.raises(ValueError):
            query_landfire(
                area_of_interest=sample_aoi,
                year=2025,
                directory=TMP_DIR,
                input_epsg=4326,
            )

    def test_assign_targets_from_sb40(self):
        sample_aoi = self.get_geojson()
        query = query_landfire(
            area_of_interest=sample_aoi, year=2019, directory=TMP_DIR, input_epsg=4326
        )
        # test just grass loading
        grass_loading = assign_targets_from_sb40(query, "grass", "loading")
        assert isinstance(grass_loading, Targets)
        assert grass_loading.method == "maxmin"
        assert grass_loading.args == ["max", "min"]
        assert len(grass_loading.targets) == 2
        # test fuel and parameter with only one value
        with pytest.warns(UserWarning):
            grass_moisture = assign_targets_from_sb40(query, "grass", "moisture")
        grass_moisture = assign_targets_from_sb40(
            query, "grass", "moisture", method="constant"
        )
        # get the rest of the fuels and params with maxmin
        grass_depth = assign_targets_from_sb40(query, "grass", "depth")
        litter_loading = assign_targets_from_sb40(query, "litter", "loading")
        litter_moisture = assign_targets_from_sb40(query, "litter", "loading")
        litter_depth = assign_targets_from_sb40(query, "litter", "loading")
        all_loading = assign_targets_from_sb40(query, "all", "loading")
        all_moisture = assign_targets_from_sb40(query, "all", "moisture")
        all_depth = assign_targets_from_sb40(query, "all", "depth")
        # test a couple with meansd
        grass_loading_meansd = assign_targets_from_sb40(
            query, "grass", "loading", method="meansd"
        )
        all_depth = assign_targets_from_sb40(query, "all", "depth", "meansd")
        # test wrong inputs
        with pytest.raises(ValueError):
            grass_depth = assign_targets_from_sb40(query, "both", "depth")
        with pytest.raises(ValueError):
            grass_depth = assign_targets_from_sb40(query, "all", "moist")
        with pytest.raises(ValueError):
            grass_depth = assign_targets_from_sb40(
                query, "all", "depth", method="minmax"
            )


def plot_array(x, title):
    plt.figure(2)
    plt.set_cmap("viridis")
    plt.imshow(x, origin="lower")
    plt.colorbar()
    plt.title(title, fontsize=18)
    plt.show()
